@@ -1,5 +1,7 @@
-#!/usr/bin/env python2
+#!/usr/bin/env python3
+
 import tarfile
+from io import TextIOWrapper
 
 unifont = 'unifont-8.0.01.tar.gz'
 loc = 'unifont-8.0.01/font/plane00/'
@@ -13,8 +15,7 @@ with open(output, 'w') as f:
     with open('header.txt') as g:
         f.write(g.read())
     for name in hexfiles:
-        z = unizip.extractfile(name)
-        f.write(z.read())
-        z.close()
+        with TextIOWrapper(unizip.extractfile(name)) as z:
+            f.write(z.read())
 
 
