@@ -1,12 +1,14 @@
 """
 PC-BASIC - GW-BASIC/BASICA/Cartridge BASIC compatible interpreter
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2020 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
-from .api import Session
-from ..metadata import VERSION as __version__
+from .data import NAME, VERSION, LONG_VERSION, AUTHOR, COPYRIGHT
+from .api import Session, codepage, font
 from .debug import DebugSession
 from .base.error import *
 from .base import signals, scancode, eascii
+
+__version__ = VERSION
