@@ -2,7 +2,7 @@
 PC-BASIC - devicebase.py
 Devices, Files and I/O operations
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2020 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -12,17 +12,18 @@ import struct
 import logging
 from contextlib import contextmanager
 
+from ...compat import iterchar
 from ..base import error
 from ..base.eascii import as_bytes as ea
 from .. import values
 
 def nullstream():
-    return open(os.devnull, b'r+')
+    return open(os.devnull, 'r+b')
 
 
 # magic chars used by some devices to indicate file type
-TYPE_TO_MAGIC = { b'B': b'\xFF', b'P': b'\xFE', b'M': b'\xFD' }
-MAGIC_TO_TYPE = { b'\xFF': b'B', b'\xFE': b'P', b'\xFD': b'M' }
+TYPE_TO_MAGIC = {b'B': b'\xFF', b'P': b'\xFE', b'M': b'\xFD'}
+MAGIC_TO_TYPE = {b'\xFF': b'B', b'\xFE': b'P', b'\xFD': b'M'}
 
 
 
@@ -57,8 +58,10 @@ class NullDevice(object):
     def __init__(self):
         """Set up device."""
 
-    def open(self, number, param, filetype, mode, access, lock,
-                   reclen, seg, offset, length, field):
+    def open(
+            self, number, param, filetype, mode, access, lock,
+            reclen, seg, offset, length, field
+        ):
         """Open a file on the device."""
         return TextFileBase(nullstream(), filetype, mode)
 
@@ -73,7 +76,7 @@ class NullDevice(object):
 class Device(object):
     """Device interface for master-file devices."""
 
-    allowed_modes = ''
+    allowed_modes = b''
 
     def __init__(self):
         """Set up device."""
@@ -81,7 +84,8 @@ class Device(object):
 
     def open(
             self, number, param, filetype, mode, access, lock,
-            reclen, seg, offset, length, field):
+            reclen, seg, offset, length, field
+        ):
         """Open a file on the device."""
         if not self.device_file:
             raise error.BASICError(error.DEVICE_UNAVAILABLE)
@@ -103,21 +107,23 @@ class Device(object):
 class SCRNDevice(Device):
     """Screen device (SCRN:) """
 
-    allowed_modes = 'OR'
+    allowed_modes = b'OR'
 
-    def __init__(self, display):
+    def __init__(self, display, console):
         """Initialise screen device."""
         # open a master file on the screen
         Device.__init__(self)
-        self.device_file = SCRNFile(display)
+        self.device_file = SCRNFile(display, console)
 
     def open(
             self, number, param, filetype, mode, access, lock,
-            reclen, seg, offset, length, field):
+            reclen, seg, offset, length, field
+        ):
         """Open a file on the device."""
         new_file = Device.open(
-                self, number, param, filetype, mode, access, lock,
-                reclen, seg, offset, length, field)
+            self, number, param, filetype, mode, access, lock,
+            reclen, seg, offset, length, field
+        )
         # SAVE "SCRN:" includes a magic byte
         new_file.write(TYPE_TO_MAGIC.get(filetype, b''))
         return new_file
@@ -126,7 +132,7 @@ class SCRNDevice(Device):
 class KYBDDevice(Device):
     """Keyboard device (KYBD:) """
 
-    allowed_modes = 'IR'
+    allowed_modes = b'IR'
 
     def __init__(self, keyboard, display):
         """Initialise keyboard device."""
@@ -201,7 +207,7 @@ class RawFile(object):
 #
 #   read_line(self)
 #   input_entry(self, typechar, allow_past_end)
-#   write_line(self, s='')
+#   write_line(self, s=b'')
 #   eof(self)
 #   lof(self)
 #   loc(self)
@@ -252,7 +258,7 @@ class TextFileBase(RawFile):
         to_read = num - len(self._readahead)
         if to_read > 0:
             with safe_io():
-                self._readahead.extend(list(self._fhandle.read(to_read)))
+                self._readahead.extend(iterchar(self._fhandle.read(to_read)))
         return b''.join(self._readahead[:num])
 
     def read(self, num):
@@ -267,7 +273,7 @@ class TextFileBase(RawFile):
         if len(output) <= 1:
             self._previous = self._current
         else:
-            self._previous = output[-2]
+            self._previous = output[-2:]
         self._current = output[-1:]
         return output
 
@@ -301,22 +307,25 @@ class TextFileBase(RawFile):
 
     def write(self, s, can_break=True):
         """Write the string s to the file, taking care of width settings."""
+        assert isinstance(s, bytes)
         # only break lines at the start of a new string. width 255 means unlimited width
         s_width = 0
         newline = False
         # find width of first line in s
-        for c in s:
+        for c in iterchar(s):
             if c in (b'\r', b'\n'):
                 newline = True
                 break
             if c >= b' ':
                 # nonprinting characters including tabs are not counted for WIDTH
                 s_width += 1
-        if (can_break and self.width != 255 and self.col != 1 and
-                self.col-1 + s_width > self.width and not newline):
+        if (
+                can_break and self.width != 255 and self.col != 1 and
+                self.col-1 + s_width > self.width and not newline
+            ):
             self.write_line()
             self.col = 1
-        for c in s:
+        for c in iterchar(s):
             # don't replace CR or LF with CRLF when writing to files
             if c == b'\r':
                 self._fhandle.write(c)
@@ -330,7 +339,7 @@ class TextFileBase(RawFile):
                     if self.col == 257:
                         self.col = 1
 
-    def write_line(self, s=''):
+    def write_line(self, s=b''):
         """Write string and follow with device-standard line break."""
         self.write(s + b'\r')
 
@@ -384,8 +393,10 @@ class InputMixin(object):
             raise error.BASICError(error.INPUT_PAST_END)
         # we read the ending char before breaking the loop
         # this may raise FIELD OVERFLOW
-        while c and not ((typechar != values.STR and c in self.soft_sep) or
-                        (c in b',\r' and not quoted)):
+        while c and not (
+                (typechar != values.STR and c in self.soft_sep) or
+                (c in b',\r' and not quoted)
+            ):
             if c == b'"' and quoted:
                 # whitespace after quote will be skipped below
                 break
@@ -417,7 +428,7 @@ class InputMixin(object):
         # skip trailing whitespace before any comma or hard separator
         if c and c in INPUT_WHITESPACE or (quoted and c == b'"'):
             self._skip_whitespace(b' ')
-            if (self.peek(1) in b',\r'):
+            if self.peek(1) in b',\r':
                 c = self.read_one()
         # file position is at one past the separator char
         return word, c
@@ -595,21 +606,21 @@ class KYBDFile(TextFileBase, RealTimeInputMixin):
 class SCRNFile(RawFile):
     """SCRN: file, allows writing to the screen as a text file."""
 
-    def __init__(self, display):
+    def __init__(self, display, console):
         """Initialise screen file."""
         RawFile.__init__(self, nullstream(), filetype=b'D', mode=b'O')
         # need display object as WIDTH can change graphics mode
         self._display = display
         # screen member is public, needed by print_
-        self.screen = display.text_screen
-        self._width = self.screen.mode.width
-        self._col = self.screen.current_col
+        self.console = console
+        self._width = self.console.width
+        self._col = self.console.current_col
         # on master-file devices, this is the master file.
         self._is_master = True
 
     def open_clone(self, filetype, mode, reclen=128):
         """Clone screen file."""
-        inst = SCRNFile(self._display)
+        inst = SCRNFile(self._display, self.console)
         inst.mode = mode
         inst.reclen = reclen
         inst.filetype = filetype
@@ -618,17 +629,19 @@ class SCRNFile(RawFile):
 
     def write(self, s, can_break=True):
         """Write string s to SCRN: """
+        if not s:
+            return
         # writes to SCRN files should *not* be echoed
         do_echo = self._is_master
-        self._col = self.screen.current_col
+        self._col = self.console.current_col
         # take column 80+overflow into account
-        if self.screen.overflow:
+        if self.console.overflow:
             self._col += 1
         # only break lines at the start of a new string. width 255 means unlimited width
         s_width = 0
         newline = False
         # find width of first line in s
-        for c in s:
+        for c in iterchar(s):
             if c in (b'\r', b'\n'):
                 newline = True
                 break
@@ -639,32 +652,37 @@ class SCRNFile(RawFile):
             elif c >= b' ':
                 # nonprinting characters including tabs are not counted for WIDTH
                 s_width += 1
-        if can_break and (self.width != 255 and self.screen.current_row != self.screen.mode.height
+        if can_break and (self.width != 255 and self.console.current_row != self.console.height
                 and self.col != 1 and self.col-1 + s_width > self.width and not newline):
-            self.screen.write_line(do_echo=do_echo)
+            self.console.write_line(do_echo=do_echo)
             self._col = 1
-        cwidth = self.screen.mode.width
-        for c in s:
+        cwidth = self.console.width
+        output = []
+        for c in iterchar(s):
             if self.width <= cwidth and self.col > self.width:
-                self.screen.write_line(do_echo=do_echo)
+                self.console.write_line(b''.join(output), do_echo=do_echo)
+                output = []
                 self._col = 1
             if self.col <= cwidth or self.width <= cwidth:
-                self.screen.write(c, do_echo=do_echo)
+                output.append(c)
             if c in (b'\n', b'\r'):
+                self.console.write(b''.join(output), do_echo=do_echo)
+                output = []
                 self._col = 1
             else:
                 self._col += 1
+        self.console.write(b''.join(output), do_echo=do_echo)
 
     def write_line(self, inp=b''):
         """Write a string to the screen and follow by CR."""
         self.write(inp)
-        self.screen.write_line(do_echo=self._is_master)
+        self.console.write_line(do_echo=self._is_master)
 
     @property
     def col(self):
         """Return current (virtual) column position."""
         if self._is_master:
-            return self.screen.current_col
+            return self.console.current_col
         else:
             return self._col
 
