@@ -2,17 +2,20 @@
 PC-BASIC - implementation.py
 Top-level implementation and main interpreter loop
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2020 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 import io
 import os
 import sys
-import Queue
+import math
 import logging
+from functools import partial
 from contextlib import contextmanager
 
-from ..metadata import NAME, VERSION, COPYRIGHT
+from ..compat import queue, text_type
+
+from .data import NAME, VERSION, COPYRIGHT
 from .base import error
 from .base import tokens as tk
 from .base import signals
@@ -23,7 +26,7 @@ from . import eventcycle
 from . import basicevents
 from . import program
 from . import display
-from . import editor
+from . import console
 from . import inputs
 from . import clock
 from . import dos
@@ -47,19 +50,20 @@ GREETING = (
 class Implementation(object):
     """Interpreter session, implementation class."""
 
-    def __init__(self,
-            syntax=u'advanced', double=False, term=u'', shell=u'',
-            output_streams=sys.stdout, input_streams=sys.stdin,
+    def __init__(
+            self, syntax=u'advanced', double=False, term=u'', shell=u'',
+            output_streams=u'stdio', input_streams=u'stdio',
             codepage=None, box_protect=True, font=None, text_width=80,
-            video=u'cga', monitor=u'rgb', aspect_ratio=(4, 3), low_intensity=False,
-            devices=None, current_device=u'Z:', mount=None, utf8=False, soft_linefeed=False,
-            keys=u'', check_keybuffer_full=True, ctrl_c_is_break=True,
+            video=u'cga', monitor=u'rgb',
+            devices=None, current_device=u'Z:',
+            textfile_encoding=None, soft_linefeed=False,
+            check_keybuffer_full=True, ctrl_c_is_break=True,
             hide_listing=None, hide_protected=False,
             peek_values=None, allow_code_poke=False, rebuild_offsets=True,
             max_memory=65534, reserved_memory=3429, video_memory=262144,
             serial_buffer_size=128, max_reclen=128, max_files=3,
-            extension=None, greeting=True,
-            ):
+            extension=()
+        ):
         """Initialise the interpreter session."""
         ######################################################################
         # session-level members
@@ -74,15 +78,14 @@ class Implementation(object):
         self._edit_prompt = False
         # terminal program for TERM command
         self._term_program = term
-        # option to suppress greeting
-        self._greeting = greeting
         ######################################################################
         # data segment
         ######################################################################
         # set up variables and memory model state
         # initialise the data segment
         self.memory = memory.DataSegment(
-                    max_memory, reserved_memory, max_reclen, max_files, double)
+            max_memory, reserved_memory, max_reclen, max_files, double
+        )
         # values and variables
         self.strings = self.memory.strings
         self.values = self.memory.values
@@ -95,8 +98,9 @@ class Implementation(object):
         # initialise the program
         bytecode = codestream.TokenisedStream(self.memory.code_start)
         self.program = program.Program(
-                self.tokeniser, self.lister, hide_listing, hide_protected,
-                allow_code_poke, self.memory, bytecode, rebuild_offsets)
+            self.tokeniser, self.lister, hide_listing, hide_protected,
+            allow_code_poke, self.memory, bytecode, rebuild_offsets
+        )
         # register all data segment users
         self.memory.set_buffers(self.program)
         ######################################################################
@@ -107,30 +111,39 @@ class Implementation(object):
         # set up input event handler
         # no interface yet; use dummy queues
         self.queues = eventcycle.EventQueues(
-                self.values, ctrl_c_is_break, inputs=Queue.Queue())
+            self.values, ctrl_c_is_break, inputs=queue.Queue()
+        )
         # prepare I/O streams
         self.io_streams = iostreams.IOStreams(
-                self.queues, self.codepage, input_streams, output_streams, utf8)
+            self.queues, self.codepage, input_streams, output_streams,
+        )
         # initialise sound queue
         self.sound = sound.Sound(self.queues, self.values, self.memory, syntax)
-        # Sound is needed for the beeps on \a
-        # InputMethods is needed for wait() in graphics
+        # initialise video
         self.display = display.Display(
-                self.queues, self.values, self.queues,
-                self.memory, text_width, video_memory, video, monitor,
-                self.sound, self.io_streams,
-                low_intensity, aspect_ratio,
-                self.codepage, font)
-        self.screen = self.display.text_screen
-        self.drawing = self.display.drawing
-        # initilise floating-point error message stream
-        self.values.set_handler(values.FloatErrorHandler(self.screen))
+            self.queues, self.values, self.queues,
+            self.memory, text_width, video_memory, video, monitor,
+            self.codepage, font
+        )
+        self.text_screen = self.display.text_screen
+        self.graphics = self.display.graphics
         # prepare input devices (keyboard, pen, joystick, clipboard-copier)
         # EventHandler needed for wait() only
         self.keyboard = inputs.Keyboard(
-                self.queues, self.values, self.codepage, keys, check_keybuffer_full)
+            self.queues, self.values, self.codepage, check_keybuffer_full
+        )
         self.pen = inputs.Pen()
         self.stick = inputs.Stick(self.values)
+        # 12 definable function keys for Tandy, 10 otherwise
+        num_fn_keys = 12 if syntax == 'tandy' else 10
+        # initialise the console
+        # Sound is needed for the beeps on \a
+        self.console = console.Console(
+            self.text_screen, self.display.cursor,
+            self.keyboard, self.sound, self.io_streams, num_fn_keys
+        )
+        # initilise floating-point error message stream
+        self.values.set_handler(values.FloatErrorHandler(self.console))
         ######################################################################
         # devices
         ######################################################################
@@ -138,15 +151,23 @@ class Implementation(object):
         # DataSegment needed for COMn and disk FIELD buffers
         # EventCycle needed for wait()
         self.files = Files(
-                self.values, self.memory, self.queues, self.keyboard, self.display,
-                max_files, max_reclen, serial_buffer_size,
-                devices, current_device, mount, utf8, not soft_linefeed)
+            self.values, self.memory, self.queues, self.keyboard, self.display, self.console,
+            max_files, max_reclen, serial_buffer_size,
+            devices, current_device,
+            self.codepage, textfile_encoding, soft_linefeed
+        )
+        # enable printer echo from console
+        self.console.set_lpt1_file(self.files.lpt1_file)
+        ######################################################################
+        # other components
+        ######################################################################
         # set up the SHELL command
         # Files needed for current disk device
         self.shell = dos.Shell(
-                self.queues, self.keyboard, self.screen, self.files, self.codepage, shell)
+            self.queues, self.keyboard, self.console, self.files, self.codepage, shell
+        )
         # set up environment
-        self.environment = dos.Environment(self.values)
+        self.environment = dos.Environment(self.values, self.codepage)
         # initialise random number generator
         self.randomiser = values.Randomiser(self.values)
         # initialise system clock
@@ -155,25 +176,17 @@ class Implementation(object):
         # register input event handlers
         ######################################################################
         # clipboard and print screen handler
-        clip_handler = inputs.ScreenCopyHandler(self.screen, self.files.lpt1_file)
-        self.queues.add_handler(clip_handler)
+        self.queues.add_handler(basicevents.ScreenCopyHandler(
+            self.queues, self.text_screen, self.files.lpt1_file
+        ))
         # keyboard, pen and stick
         self.queues.add_handler(self.keyboard)
         self.queues.add_handler(self.pen)
         self.queues.add_handler(self.stick)
         # set up BASIC event handlers
         self.basic_events = basicevents.BasicEvents(
-                self.values, self.sound, self.clock, self.files,
-                self.screen, self.program, syntax)
-        ######################################################################
-        # editor
-        ######################################################################
-        # key macro guide
-        self.fkey_macros = editor.FunctionKeyMacros(
-                self.keyboard, self.screen, self.basic_events.num_fn_keys)
-        # initialise the editor
-        self.editor = editor.Editor(
-                self.screen, self.keyboard, self.sound, self.io_streams, self.files.lpt1_file)
+            self.values, self.sound, self.clock, self.files, self.program, num_fn_keys
+        )
         ######################################################################
         # extensions
         ######################################################################
@@ -185,19 +198,22 @@ class Implementation(object):
         self.parser = parser.Parser(self.values, self.memory, syntax)
         # initialise the interpreter
         self.interpreter = interpreter.Interpreter(
-                self.queues, self.screen, self.files, self.sound,
-                self.values, self.memory, self.program, self.parser, self.basic_events)
+            self.queues, self.console, self.display.cursor, self.files, self.sound,
+            self.values, self.memory, self.program, self.parser, self.basic_events
+        )
         ######################################################################
         # callbacks
         ######################################################################
         # set up non-data segment memory
         self.all_memory = machine.Memory(
-                self.values, self.memory, self.files,
-                self.display, self.keyboard, self.screen.fonts[8],
-                self.interpreter, peek_values, syntax)
+            self.values, self.memory, self.files,
+            self.display, self.keyboard, self.display.memory_font,
+            self.interpreter, peek_values, syntax
+        )
         # initialise machine ports
         self.machine = machine.MachinePorts(
-                self.queues, self.values, self.display, self.keyboard, self.stick, self.files)
+            self.queues, self.values, self.display, self.keyboard, self.stick, self.files
+        )
         # build function table (depends on Memory having been initialised)
         self.parser.init_callbacks(self)
 
@@ -227,12 +243,10 @@ class Implementation(object):
         else:
             # use dummy video & audio queues if not provided
             # but an input queue should be operational for I/O streams
-            self.queues.set(inputs=Queue.Queue())
+            self.queues.set(inputs=queue.Queue())
 
     def execute(self, command):
         """Execute a BASIC statement."""
-        # don't greet if an interactive session is opened afterwards
-        self._greeting = False
         with self._handle_exceptions():
             self._store_line(command)
             self.interpreter.loop()
@@ -250,30 +264,52 @@ class Implementation(object):
     def set_variable(self, name, value):
         """Set a variable in memory."""
         name = name.upper()
-        if isinstance(value, unicode):
-            value = self.codepage.str_from_unicode(value)
+        if isinstance(value, text_type):
+            value = self.codepage.unicode_to_bytes(value)
         elif isinstance(value, bool):
             value = -1 if value else 0
         if b'(' in name:
             name = name.split(b'(', 1)[0]
             self.arrays.from_list(value, name)
         else:
-            self.memory.set_variable(name, [], self.values.from_value(value, name[-1]))
+            self.memory.set_variable(name, [], self.values.from_value(value, name[-1:]))
 
-    def get_variable(self, name):
+    def get_converter(self, from_type, to_type):
+        """Get a converter function; raise ValueError if not allowed"""
+        if to_type is None or from_type == to_type:
+            return lambda _x: _x
+        converter = {
+            (bytes, text_type): partial(self.codepage.bytes_to_unicode, preserve=cp.CONTROL),
+            (text_type, bytes): self.codepage.unicode_to_bytes,
+            (int, bool): bool,
+            (float, bool): bool,
+            (bool, int): lambda _bool: (-1 if _bool else 0),
+            (int, float): float,
+            (float, int): lambda _flt: int(math.floor(_flt)),
+            (bool, float): lambda _bool: (-1. if _bool else 0.),
+        }
+        try:
+            return converter[(from_type, to_type)]
+        except KeyError:
+            raise ValueError("BASIC can't convert %s to %s." % (from_type, to_type))
+
+    def get_variable(self, name, as_type=None):
         """Get a variable in memory."""
         name = name.upper()
-        if '(' in name:
-            name = name.split('(', 1)[0]
-            return self.arrays.to_list(name)
+        if b'(' in name:
+            name = name.split(b'(', 1)[0]
+            value = self.arrays.to_list(name)
+            if not value:
+                return []
+            convert = self.get_converter(type(value[0]), as_type)
+            return [convert(_item) for _item in value]
         else:
-            return self.memory.view_or_create_variable(name, []).to_value()
+            value = self.memory.view_or_create_variable(name, []).to_value()
+            convert = self.get_converter(type(value), as_type)
+            return convert(value)
 
     def interact(self):
         """Interactive interpreter session."""
-        # greet at most once per session: execute() will switch off greeting
-        if self._greeting:
-            self.execute(GREETING)
         while True:
             with self._handle_exceptions():
                 self.interpreter.loop()
@@ -282,7 +318,7 @@ class Implementation(object):
                 else:
                     self._show_prompt()
                     # input loop, checks events
-                    line = self.editor.wait_screenline(from_start=True)
+                    line = self.console.read_line(from_start=True)
                     self._prompt = not self._store_line(line)
 
     def close(self):
@@ -290,6 +326,8 @@ class Implementation(object):
         # close files if we opened any
         self.files.close_all()
         self.files.close_devices()
+        # kill the iostreams threads so windows doesn't run out
+        self.io_streams.close()
 
     def _show_prompt(self):
         """Show the Ok or EDIT prompt, unless suppressed."""
@@ -297,10 +335,10 @@ class Implementation(object):
             linenum, tell = self._edit_prompt
             # unset edit prompt first, in case program.edit throws
             self._edit_prompt = False
-            self.program.edit(self.screen, linenum, tell)
+            self.program.edit(self.console, linenum, tell)
         elif self._prompt:
-            self.screen.start_line()
-            self.screen.write_line('Ok\xff')
+            self.console.start_line()
+            self.console.write_line(b'Ok\xff')
 
     def _store_line(self, line):
         """Store a program line or schedule a command line for execution."""
@@ -308,7 +346,7 @@ class Implementation(object):
             return True
         self.interpreter.direct_line = self.tokeniser.tokenise_line(line)
         c = self.interpreter.direct_line.peek()
-        if c == '\0':
+        if c == b'\0':
             # clear all program stacks
             self.interpreter.clear_stacks_and_pointers()
             # clear variables first,
@@ -318,7 +356,7 @@ class Implementation(object):
             self.program.check_number_start(self.interpreter.direct_line)
             self.program.store_line(self.interpreter.direct_line)
             return True
-        elif c != '':
+        elif c != b'':
             # it is a command, go and execute
             self.interpreter.set_parse_mode(True)
             return False
@@ -326,20 +364,19 @@ class Implementation(object):
     def _auto_step(self):
         """Generate an AUTO line number and wait for input."""
         try:
-            numstr = str(self._auto_linenum)
-            self.screen.write(numstr)
+            numstr = b'%d' % (self._auto_linenum,)
             if self._auto_linenum in self.program.line_numbers:
-                self.screen.write('*')
-                line = bytearray(self.editor.wait_screenline(from_start=True))
-                if line[:len(numstr)+1] == numstr + '*':
-                    line[len(numstr)] = ' '
+                prompt = numstr + b'*'
             else:
-                self.screen.write(' ')
-                line = bytearray(self.editor.wait_screenline(from_start=True))
+                prompt = numstr + b' '
+            line = self.console.read_line(prompt, from_start=True)
+            # remove *, if present
+            if line[:len(numstr)+1] == b'%s*' % (numstr,):
+                line = b'%s %s' % (numstr, line[len(numstr)+1:])
             # run or store it; don't clear lines or raise undefined line number
             self.interpreter.direct_line = self.tokeniser.tokenise_line(line)
             c = self.interpreter.direct_line.peek()
-            if c == '\0':
+            if c == b'\0':
                 # check for lines starting with numbers (6553 6) and empty lines
                 empty, scanline = self.program.check_number_start(self.interpreter.direct_line)
                 if not empty:
@@ -348,7 +385,7 @@ class Implementation(object):
                     self.interpreter.clear_stacks_and_pointers()
                     self._clear_all()
                 self._auto_linenum = scanline + self._auto_increment
-            elif c != '':
+            elif c != b'':
                 # it is a command, go and execute
                 self.interpreter.set_parse_mode(True)
         except error.Break:
@@ -377,8 +414,8 @@ class Implementation(object):
     def _handle_error(self, e):
         """Handle a BASIC error through error message."""
         # not handled by ON ERROR, stop execution
-        self.screen.start_line()
-        self.screen.write(e.get_message(self.program.get_line_number(e.pos)))
+        self.console.start_line()
+        self.console.write(e.get_message(self.program.get_line_number(e.pos)))
         self.interpreter.set_parse_mode(False)
         self.interpreter.input_mode = False
         # special case: syntax error
@@ -419,7 +456,7 @@ class Implementation(object):
             # select video memory size (Tandy/PCjr only)
             video_size = next(args)
             if video_size is not None:
-                video_size = values.round(video_size).to_value()
+                video_size = round(video_size.to_value())
                 self.display.set_video_memory_size(video_size)
             # execute any remaining parsing steps
             next(args)
@@ -446,7 +483,7 @@ class Implementation(object):
         # reset PLAY state
         self.sound.reset_play()
         # reset DRAW state (angle, scale) and current graphics position
-        self.drawing.reset()
+        self.graphics.reset()
         # reset random number generator
         self.randomiser.clear()
         # reset stacks & pointers
@@ -456,14 +493,14 @@ class Implementation(object):
         """SHELL: open OS shell and optionally execute command."""
         cmd = values.next_string(args)
         list(args)
-        # force cursor visible in all cases
-        self.screen.cursor.show(True)
+        # force cursor visible
+        self.display.cursor.set_override(True)
         # sound stops playing and is forgotten
         self.sound.stop_all_sound()
         # run the os-specific shell
         self.shell.launch(cmd)
         # reset cursor visibility to its previous state
-        self.screen.cursor.reset_visibility()
+        self.display.cursor.set_override(False)
 
     def term_(self, args):
         """TERM: terminal emulator."""
@@ -476,7 +513,7 @@ class Implementation(object):
             raise error.BASICError(error.INTERNAL_ERROR)
         # terminal program for TERM command
         prog = self.files.get_device(b'@:').bind(self._term_program)
-        with self.files.open(0, prog, filetype='ABP', mode='I') as progfile:
+        with self.files.open(0, prog, filetype=b'ABP', mode=b'I') as progfile:
             self.program.load(progfile)
         self.interpreter.error_handle_mode = False
         self.interpreter.clear_stacks_and_pointers()
@@ -497,7 +534,7 @@ class Implementation(object):
         line_range = next(args)
         out = values.next_string(args)
         if out is not None:
-            out = self.files.open(0, out, filetype='A', mode='O')
+            out = self.files.open(0, out, filetype=b'A', mode=b'O')
         list(args)
         lines = self.program.list_lines(*line_range)
         if out:
@@ -510,7 +547,7 @@ class Implementation(object):
                 # and interruptible
                 self.queues.wait()
                 # LIST on screen is slightly different from just writing
-                self.screen.list_line(l)
+                self.console.list_line(l)
         # return to direct mode
         self.interpreter.set_pointer(False)
 
@@ -524,7 +561,6 @@ class Implementation(object):
         # throws back to direct mode
         # jump to end of direct line so execution stops
         self.interpreter.set_pointer(False)
-        self.screen.cursor.reset_visibility()
         # request edit prompt
         self._edit_prompt = (from_line, None)
 
@@ -544,7 +580,7 @@ class Implementation(object):
         """LOAD: load program from file."""
         name = values.next_string(args)
         comma_r, = args
-        with self.files.open(0, name, filetype='ABP', mode='I') as f:
+        with self.files.open(0, name, filetype=b'ABP', mode=b'I') as f:
             self.program.load(f)
         # reset stacks
         self.interpreter.clear_stacks_and_pointers()
@@ -565,7 +601,7 @@ class Implementation(object):
         if jumpnum is not None:
             jumpnum = values.to_int(jumpnum, unsigned=True)
         common_all, delete_lines = next(args), next(args)
-        from_line, to_line = delete_lines if delete_lines else None, None
+        from_line, to_line = delete_lines if delete_lines else (None, None)
         if to_line is not None and to_line not in self.program.line_numbers:
             raise error.BASICError(error.IFC)
         list(args)
@@ -582,7 +618,7 @@ class Implementation(object):
                     preserve_base=(commons or common_all),
                     preserve_deftype=merge)
             # load new program
-            with self.files.open(0, name, filetype='ABP', mode='I') as f:
+            with self.files.open(0, name, filetype=b'ABP', mode=b'I') as f:
                 if delete_lines:
                     # delete lines from existing code before merge
                     # (without MERGE, this is pointless)
@@ -603,13 +639,15 @@ class Implementation(object):
     def save_(self, args):
         """SAVE: save program to a file."""
         name = values.next_string(args)
-        mode = (next(args) or 'B').upper()
+        mode = (next(args) or b'B').upper()
         list(args)
-        with self.files.open(0, name, filetype=mode, mode='O',
-                            seg=self.memory.data_segment, offset=self.memory.code_start,
-                            length=len(self.program.bytecode.getvalue())-1) as f:
+        with self.files.open(
+                0, name, filetype=mode, mode=b'O',
+                seg=self.memory.data_segment, offset=self.memory.code_start,
+                length=len(self.program.bytecode.getvalue())-1
+            ) as f:
             self.program.save(f)
-        if mode == 'A':
+        if mode == b'A':
             # return to direct mode
             self.interpreter.set_pointer(False)
 
@@ -618,7 +656,7 @@ class Implementation(object):
         name = values.next_string(args)
         list(args)
         # check if file exists, make some guesses (all uppercase, +.BAS) if not
-        with self.files.open(0, name, filetype='A', mode='I') as f:
+        with self.files.open(0, name, filetype=b'A', mode=b'I') as f:
             self.program.merge(f)
         # clear all program stacks
         self.interpreter.clear_stacks_and_pointers()
@@ -643,7 +681,7 @@ class Implementation(object):
             try:
                 name = values.next_string(args)
                 comma_r = next(args)
-                with self.files.open(0, name, filetype='ABP', mode='I') as f:
+                with self.files.open(0, name, filetype=b'ABP', mode=b'I') as f:
                     self.program.load(f)
             except StopIteration:
                 pass
@@ -675,7 +713,7 @@ class Implementation(object):
         if file_number is not None:
             file_number = values.to_int(file_number)
             error.range_check(0, 255, file_number)
-            finp = self.files.get(file_number, mode='IR')
+            finp = self.files.get(file_number, mode=b'IR')
             self._input_file(finp, args)
         else:
             newline, prompt, following = next(args)
@@ -683,57 +721,53 @@ class Implementation(object):
 
     def _input_console(self, newline, prompt, following, readvar):
         """INPUT: request input from user."""
-        if following == ';':
-            prompt += '? '
-        # read the input
-        self.interpreter.input_mode = True
-        self.parser.redo_on_break = True
-        # readvar is a list of (name, indices) tuples
-        # we return a list of (name, indices, values) tuples
-        while True:
-            self.screen.write(prompt)
-            # disconnect the wrap between line with the prompt and previous line
-            if self.screen.current_row > 1:
-                self.screen.text.pages[self.screen.apagenum].row[self.screen.current_row-2].wrap = False
-            line = self.editor.wait_screenline(write_endl=newline)
-            inputstream = InputTextFile(line)
-            # read the values and group them and the separators
-            var, values, seps = [], [], []
-            for name, indices in readvar:
-                name = self.memory.complete_name(name)
-                word, sep = inputstream.input_entry(name[-1], allow_past_end=True)
-                try:
-                    value = self.values.from_repr(word, allow_nonnum=False, typechar=name[-1])
-                except error.BASICError as e:
-                    # string entered into numeric field
-                    value = None
-                var.append([name, indices])
-                values.append(value)
-                seps.append(sep)
-            # last separator not empty: there were too many values or commas
-            # earlier separators empty: there were too few values
-            # empty values will be converted to zero by from_str
-            # None means a conversion error occurred
-            if (seps[-1] or '' in seps[:-1] or None in values):
-                # good old Redo!
-                self.screen.write_line('?Redo from start')
-                readvar = var
-            else:
-                varlist = [r + [v] for r, v in zip(var, values)]
-                break
-        self.parser.redo_on_break = False
-        self.interpreter.input_mode = False
-        for v in varlist:
-            self.memory.set_variable(*v)
+        if following == b';':
+            prompt += b'? '
+        with self.memory.get_stack() as stack:
+            # read the input
+            self.interpreter.input_mode = True
+            self.parser.redo_on_break = True
+            # readvar is a list of (name, indices) tuples
+            # we return a list of (name, indices, values) tuples
+            while True:
+                line = self.console.read_line(prompt, write_endl=newline)
+                inputstream = InputTextFile(line)
+                # read the values and group them and the separators
+                var, values, seps = [], [], []
+                for name, indices in readvar:
+                    name = self.memory.complete_name(name)
+                    word, sep = inputstream.input_entry(name[-1:], allow_past_end=True)
+                    try:
+                        value = self.values.from_repr(word, allow_nonnum=False, typechar=name[-1:])
+                    except error.BASICError as e:
+                        # string entered into numeric field
+                        value = None
+                    stack.append(value)
+                    var.append([name, indices])
+                    values.append(value)
+                    seps.append(sep)
+                # last separator not empty: there were too many values or commas
+                # earlier separators empty: there were too few values
+                # empty values will be converted to zero by from_str
+                # None means a conversion error occurred
+                if (seps[-1] or b'' in seps[:-1] or None in values):
+                    # good old Redo!
+                    self.console.write_line(b'?Redo from start')
+                    readvar = var
+                else:
+                    varlist = [r + [v] for r, v in zip(var, values)]
+                    break
+            self.parser.redo_on_break = False
+            self.interpreter.input_mode = False
+            for v in varlist:
+                self.memory.set_variable(*v)
 
     def _input_file(self, finp, readvar):
         """INPUT: retrieve input from file."""
         for v in readvar:
             name, indices = v
-            word, _ = finp.input_entry(name[-1], allow_past_end=False)
-            value = self.values.from_repr(word, allow_nonnum=True, typechar=name[-1])
-            if value is None:
-                value = self.values.new(name[-1])
+            word, _ = finp.input_entry(name[-1:], allow_past_end=False)
+            value = self.values.from_repr(word, allow_nonnum=True, typechar=name[-1:])
             self.memory.set_variable(name, indices, value)
 
     def line_input_(self, args):
@@ -747,14 +781,12 @@ class Implementation(object):
             prompt, newline = None, None
             file_number = values.to_int(file_number)
             error.range_check(0, 255, file_number)
-            finp = self.files.get(file_number, mode='IR')
+            finp = self.files.get(file_number, mode=b'IR')
         # get string variable
         readvar, indices = next(args)
         list(args)
-        if not readvar:
-            raise error.BASICError(error.STX)
         readvar = self.memory.complete_name(readvar)
-        if readvar[-1] != values.STR:
+        if readvar[-1:] != values.STR:
             raise error.BASICError(error.TYPE_MISMATCH)
         # read the input
         if finp:
@@ -764,8 +796,7 @@ class Implementation(object):
         else:
             self.interpreter.input_mode = True
             self.parser.redo_on_break = True
-            self.screen.write(prompt)
-            line = self.editor.wait_screenline(write_endl=newline)
+            line = self.console.read_line(prompt, write_endl=newline)
             self.parser.redo_on_break = False
             self.interpreter.input_mode = False
         self.memory.set_variable(readvar, indices, self.values.from_value(line, values.STR))
@@ -778,32 +809,33 @@ class Implementation(object):
             val = values.pass_number(val, err=error.IFC)
         else:
             # prompt for random seed if not specified
-            while val is None:
-                self.screen.write('Random number seed (-32768 to 32767)? ')
-                seed = self.editor.wait_screenline()
-                val = self.values.from_repr(seed, allow_nonnum=False)
+            while True:
+                seed = self.console.read_line(b'Random number seed (-32768 to 32767)? ')
+                try:
+                    val = self.values.from_repr(seed, allow_nonnum=False)
+                except error.BASICError as e:
+                    if e.err != error.IFC:
+                        raise
+                else:
+                    break
             # seed entered on prompt is rounded to int
             val = values.to_integer(val)
         self.randomiser.reseed(val)
 
     def key_(self, args):
-        """KEY: macro or event handler definition."""
+        """KEY: macro or event trigger definition."""
         keynum = values.to_int(next(args))
         error.range_check(1, 255, keynum)
         text = values.next_string(args)
         list(args)
-        if keynum <= self.basic_events.num_fn_keys:
-            self.fkey_macros.set(keynum, text)
-            self.screen.bottom_bar.redraw(self.screen)
-        else:
-            # only length-2 expressions can be assigned to KEYs over 10
-            # in which case it's a key scancode definition
-            if len(text) != 2:
-                raise error.BASICError(error.IFC)
-            self.basic_events.key[keynum-1].set_trigger(str(text))
+        try:
+            self.console.set_macro(keynum, text)
+        except ValueError:
+            # if out of range of number of macros (12 on Tandy, else 10), it's a trigger definition
+            self.basic_events.key[keynum-1].set_trigger(text)
 
     def pen_fn_(self, args):
         """PEN: poll the light pen."""
         fn, = args
-        result = self.pen.poll(fn, self.basic_events.pen.enabled, self.screen)
+        result = self.pen.poll(fn, self.basic_events.pen.enabled, self.display.apage)
         return self.values.new_integer().from_int(result)
