@@ -2,14 +2,14 @@
 PC-BASIC - mlparser.py
 DRAW and PLAY macro language stream utilities
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2020 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
-import string
-
+from ..compat import iterchar
 from .base import error
 from .base import codestream
+from .base.tokens import DIGITS
 from . import values
 
 
@@ -17,7 +17,7 @@ class MLParser(codestream.CodeStream):
     """Macro Language parser."""
 
     # whitespace character for both macro languages is only space
-    blanks = ' '
+    blanks = b' '
 
     def __init__(self, gml, data_memory, values):
         """Initialise macro-language parser."""
@@ -28,31 +28,31 @@ class MLParser(codestream.CodeStream):
     def parse_number(self, default=None):
         """Parse a value in a macro-language string."""
         c = self.skip_blank()
-        sgn = -1 if c == '-' else 1
-        if c in ('+', '-'):
+        sgn = -1 if c == b'-' else 1
+        if c in (b'+', b'-'):
             self.read(1)
             c = self.peek()
             # don't allow default if sign is given
             default = None
-        if c == '=':
+        if c == b'=':
             self.read(1)
             c = self.peek()
             if len(c) == 0:
                 raise error.BASICError(error.IFC)
             elif ord(c) > 8:
                 step = self._parse_variable().to_int()
-                self.require_read((';',), err=error.IFC)
+                self.require_read((b';',), err=error.IFC)
             else:
                 # varptr$
                 step = self.memory.get_value_for_varptrstr(self.read(3)).to_int()
-        elif c and c in string.digits:
+        elif c and c in DIGITS:
             step = self._parse_const()
         elif default is not None:
             step = default
         else:
             raise error.BASICError(error.IFC)
         if sgn == -1:
-            step = -step
+            step = -step  # pylint: disable=invalid-unary-operand-type
         return step
 
     def parse_string(self):
@@ -62,7 +62,7 @@ class MLParser(codestream.CodeStream):
             raise error.BASICError(error.IFC)
         elif ord(c) > 8:
             sub = self._parse_variable()
-            self.require_read((';',), err=error.IFC)
+            self.require_read((b';',), err=error.IFC)
             return values.pass_string(sub).to_str()
         else:
             # varptr$
@@ -78,8 +78,8 @@ class MLParser(codestream.CodeStream):
 
     def _parse_const(self):
         """Parse and return a constant value in a macro-language string."""
-        numstr = ''
-        while self.skip_blank() in set(string.digits):
+        numstr = b''
+        while self.skip_blank() in set(iterchar(DIGITS)):
             numstr += self.read(1)
         try:
             return int(numstr)
@@ -89,13 +89,13 @@ class MLParser(codestream.CodeStream):
     def _parse_indices(self):
         """Parse constant array indices."""
         indices = []
-        if self.skip_blank_read_if(('[', '(')):
+        if self.skip_blank_read_if((b'[', b'(')):
             while True:
-                if self.skip_blank() in set(string.digits):
+                if self.skip_blank() in set(iterchar(DIGITS)):
                     indices.append(self._parse_const())
                 else:
                     indices.append(self._parse_variable().to_int())
-                if not self.skip_blank_read_if((',',)):
+                if not self.skip_blank_read_if((b',',)):
                     break
-            self.require_read((']', ')'))
+            self.require_read((b']', b')'))
         return indices
