@@ -2,7 +2,7 @@
 PC-BASIC - program.py
 Program buffer utilities
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2020 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -78,6 +78,7 @@ class Program(object):
         self.line_numbers = {65536: 0}
         self.last_stored = None
         self.code_size = self.bytecode.tell()
+        self.bytecode.truncate()
 
     def truncate(self, rest=b''):
         """Write bytecode and cut the program of beyond the current position."""
@@ -93,6 +94,8 @@ class Program(object):
     def get_line_number(self, pos):
         """Get line number for stream position."""
         pre = -1
+        if pos is None:
+            pos = -1
         for linum in self.line_numbers:
             linum_pos = self.line_numbers[linum]
             if linum_pos <= pos and linum > pre:
@@ -234,16 +237,16 @@ class Program(object):
         # update line number dict
         self.update_line_dict(startpos, afterpos, 0, deleteable, beyond)
 
-    def edit(self, screen, from_line, bytepos=None):
+    def edit(self, console, from_line, bytepos=None):
         """Output program line to console and position cursor."""
         if self.protected:
-            screen.write(b'%d\r' % (from_line,))
+            console.write(b'%d\r' % (from_line,))
             raise error.BASICError(error.IFC)
         # list line
         self.bytecode.seek(self.line_numbers[from_line]+1)
         _, output, textpos = self.lister.detokenise_line(self.bytecode, bytepos)
         # no newline to avoid scrolling on line 24
-        screen.list_line(bytes(output), newline=False)
+        console.list_line(bytes(output), newline=False)
         # find row, column position for textpos
         newlines, c = 0, 0
         pos_row, pos_col = 0, 0
@@ -251,7 +254,7 @@ class Program(object):
             return
         for i, byte in enumerate(output):
             c += 1
-            if byte == ord(b'\n') or c > screen.mode.width:
+            if byte == ord(b'\n') or c > console.width:
                 newlines += 1
                 c = 0
             if i == textpos:
@@ -259,11 +262,11 @@ class Program(object):
         if textpos > i:
             pos_row, pos_col = newlines, c + 1
         if bytepos:
-            screen.set_pos(screen.current_row-newlines+pos_row, pos_col)
+            console.set_pos(console.current_row-newlines+pos_row, pos_col)
         else:
-            screen.set_pos(screen.current_row-newlines, 1)
+            console.set_pos(console.current_row-newlines, 1)
 
-    def renum(self, screen, new_line, start_line, step):
+    def renum(self, console, new_line, start_line, step):
         """Renumber stored program."""
         new_line = 10 if new_line is None else new_line
         start_line = 0 if start_line is None else start_line
@@ -311,7 +314,7 @@ class Program(object):
                 # not redefined, exists in program?
                 if jumpnum not in self.line_numbers:
                     linum = self.get_line_number(ins.tell()-1)
-                    screen.write_line(b'Undefined line %d in %d' % (jumpnum, linum))
+                    console.write_line(b'Undefined line %d in %d' % (jumpnum, linum))
                 newjump = jumpnum
             ins.seek(-2, 1)
             ins.write(struct.pack('<H', newjump))
