@@ -2,18 +2,18 @@
 PC-BASIC - sound.py
 Sound handling
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2020 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 from collections import deque
-import Queue
-import string
 import datetime
 
+from ..compat import iterchar, zip
 from .base import error
 from .base import signals
 from .base import tokens as tk
+from .base.tokens import DIGITS
 from . import mlparser
 from . import values
 
@@ -60,7 +60,7 @@ class Sound(object):
         # frequency for noise sources
         self._noise_freq = list(NOISE_FREQ)
         # advnced sound capabilities
-        self._multivoice = syntax if syntax in ('pcjr', 'tandy') else ''
+        self._multivoice = syntax if syntax in ('pcjr', 'tandy') else b''
         # Tandy/PCjr SOUND ON and BEEP ON
         # tandy has SOUND ON by default, pcjr has it OFF
         self._sound_on = (self._multivoice == 'tandy')
@@ -73,6 +73,11 @@ class Sound(object):
         # initialise PLAY state
         self.reset_play()
 
+    @property
+    def multivoice(self):
+        """We have multivoice capability."""
+        return bool(self._multivoice)
+
     def beep_(self, args):
         """BEEP: produce an alert sound or switch internal speaker on/off."""
         command, = args
@@ -220,7 +225,7 @@ class Sound(object):
         """Rebuild tone queues."""
         # should we pop one at a time from each voice queue to equalise timings?
         for voice, q in enumerate(self._voice_queue):
-            for item, duration in q.iteritems():
+            for item, duration in q.items():
                 item.params = list(item.params)
                 item.params[2] = duration
                 self._queues.audio.put(item)
@@ -263,7 +268,7 @@ class Sound(object):
     def play_(self,  args):
         """Parse a list of Music Macro Language strings (PLAY statement)."""
         # retrieve Music Macro Language string
-        mml_list = list(values.next_string(args) for _ in range(3))
+        mml_list = [values.to_string_or_none(arg) for arg, _ in zip(args, range(3))]
         list(args)
         # at least one string must be specified
         if not any(mml_list):
@@ -277,7 +282,7 @@ class Sound(object):
         mml_list += [b''] * (3 - len(mml_list))
         ml_parser_list = [mlparser.MLParser(mml, self._memory, self._values) for mml in mml_list]
         next_oct = 0
-        voices = range(3)
+        voices = list(range(3))
         while True:
             if not voices:
                 break
@@ -341,10 +346,10 @@ class Sound(object):
                         note += b'#'
                     elif mmls.skip_blank_read_if((b'-',)):
                         note += b'-'
-                    c = mmls.skip_blank_read_if(string.digits)
+                    c = mmls.skip_blank_read_if(DIGITS)
                     if c is not None:
                         numstr = [c]
-                        while mmls.skip_blank() in set(string.digits):
+                        while mmls.skip_blank() in set(iterchar(DIGITS)):
                             numstr.append(mmls.read(1))
                         # NOT ml_parse_number, only literals allowed here!
                         length = int(b''.join(numstr))
@@ -496,7 +501,7 @@ class TimedQueue(object):
         except IndexError:
             return datetime.datetime.now()
 
-    def iteritems(self):
+    def items(self):
         """Iterate over each item and its duration."""
         self._check_expired()
         last_expiry = datetime.datetime.now()
