@@ -2,16 +2,20 @@
 PC-BASIC - compat.base
 Cross-platform compatibility utilities
 
-(c) 2018 Rob Hagemans
+(c) 2018--2020 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 import os
 import re
+import contextlib
 import sys
 import platform
 
 
+# Python major version
+PY2 = sys.version_info.major == 2
+
 # platform constants
 WIN32 = sys.platform == 'win32'
 MACOS = sys.platform == 'darwin'
@@ -48,10 +52,41 @@ else:
     BASE_DIR = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))
 
 
-# utility function, this has to go somewhere...
+
+# utility functions, this has to go somewhere...
+
 def split_quoted(line, split_by=u'\s', quote=u'"', strip_quotes=False):
     """Split by separators, preserving quoted blocks."""
-    chunks = re.findall(ur'[^%s%s][^%s]*|%s.+?%s' % (quote, split_by, split_by, quote, quote), line)
+    # https://stackoverflow.com/questions/16710076/python-split-a-string-respect-and-preserve-quotes
+    regexp = r'(?:[^{split_by}{quote}]|{quote}(?:\\.|[^{quote}])*{quote})+'.format(
+        split_by=split_by, quote=quote
+    )
+    chunks = re.findall(regexp, line)
     if strip_quotes:
         chunks = [c.strip(quote) for c in chunks]
     return chunks
+
+def split_pair(s, sep):
+    """Split an argument by separator, always return two elements."""
+    slist = s.split(sep, 1)
+    s0 = slist[0]
+    if len(slist) > 1:
+        s1 = slist[1]
+    else:
+        s1 = u''
+    return s0, s1
+
+def iter_chunks(char_list, attrs):
+    """Iterate over list yielding chunks of elements with the same attribute."""
+    last_attr = None
+    chars = []
+    # collect chars in chunks with the same attribute
+    for char, attr in zip(char_list, attrs):
+        if attr != last_attr:
+            if last_attr is not None:
+                yield chars, last_attr
+            last_attr = attr
+            chars = []
+        chars.append(char)
+    if chars:
+        yield chars, attr
