@@ -2,7 +2,7 @@
 PC-BASIC - audio_beep.py
 Sound implementation through the linux beep utility
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2020 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -15,10 +15,10 @@ import sys
 from ..compat import WIN32, which
 
 if WIN32:
-    import winsound
+    import winsound  # pylint: disable=import-error
     fcntl = None
 else:
-    import fcntl
+    import fcntl  # pylint: disable=import-error
     winsound = None
 
 from .audio import AudioPlugin
@@ -167,7 +167,7 @@ class LinuxBeeper(WinBeeper):
         if frequency < 37 or frequency >= 32767:
             fcntl.ioctl(sys.stdout, KIOCSOUND, 0)
         else:
-            fcntl.ioctl(sys.stdout, KIOCSOUND, int(CLOCK_TICK_RATE / frequency))
+            fcntl.ioctl(sys.stdout, KIOCSOUND, int(CLOCK_TICK_RATE // frequency))
         time.sleep(duration)
         if not loop:
             fcntl.ioctl(sys.stdout, KIOCSOUND, 0)
